#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

/*This code simulates the Incoherent Feed-Forward Loop (IFFL) dynamics described in the paper
 using the Gillespie's algorithm. Parameters are defined as in the Main Text and given as input.
 To obtain the analysis over a wide range of some paremeters, an input scheme is implemented
 into the Main function using the command line arguments (int argc, char **argv) and can be modified as needed. The whole run over the range of chosen parameters can be performed through a shell script included in the head folder. Note that to perform one run only, it is enough to fix all parameters in this code and comment the lines concerning the command line arguments.
 
 Structure of the code
 ---------------------
 This code is made of two functions:
 1) Main - where input parameters are defined and the Gillespie's run in time is performed;
 2) Gillespie's algorithm - onestepgillespie - function performing one time step of the Gillespie's algorithm, i.e., computing the number of molecules for each species after one Gillespie's step and returning the overall reaction time.*/

/*struct for the molecular species*/
struct algorithm{
	double MTF, TF, mu, MP, c, p;  
};

struct algorithm y;
struct algorithm y2, y3, y4;


/*reaction parameters as global variables*/

double bMTF=0.005;//=0.025; /*nM/min*/
double dMTF=0.0243291; /*1/min*/
double tTF=0.01; /*1/min*/
double dTF=0.01; /*1/min*/
double tmu;//=0.0798071; /*1/min*/
double dmu=0.0154351; /*1/min*/
double bmu=0.;
double a=0.;     /*1/min*/
double kp=0.01; /*1/(nM*min)*/
double km=0.01; /*1/min*/
double bMP=0.;
double dMP=0.0243291; /*1/min*/
double tMP=0.0798071; /*1/min*/
double dc=0.; /*null probability of complex's death*/
double dp=0.01; /*1/min*/
double tp=0.0271314; /*1/min*/
double convfact=2521.21; /*number of molecules in 1nM of solution*/



int N=10000000;
double t1=0.;

double onestepgillespie(double *);

FILE *fp;
FILE *fs;
FILE *ft;

int main(int argc, char** argv) {
	double end, start, t=0.;
	long long int i=0, nsteps;
	long int tau=0;
	int h, min, sec, elapsed_time, j;
	double CVp=0., CVTF=0., phip;
	double averTF, averp, varTF, varp, sumsqTF=0., sumsqp=0., sumTF=0., sump=0., sigmap, sigmaTF, avermu, summu=0., xiTF, xip, sum, sumMTF=0.;
    double sumtri=0., sumMP=0., sumMPmu=0., sumc=0., aversumtri=0., averc=0., averMP=0., averMPmu=0., CM, CM2, varc, varMP, varmu, sumsqmu=0., sumsqMP=0., sumsqc=0., averMTF,told=0.,tnew,step=pow(10.,3.);
	
    /*define hereby the parameters you wish to vary in your simulation*/
	for(i=0; i<argc; i++){
		
		if(strstr(argv[i], "-tmu")) tmu=atof(argv[i]+4);
		
	}
	
	printf("\n\nRunning with free parameter: tmu=%lf\n\n",tmu);
	
    /*Three files will be produced: one with the results in time of Gillespie's iterations, another with the coefficient of variations of the proteins and the transcription factors and a final one with proteins number at the steady state for histograms*/
	fp=fopen("resgillespie.txt","w");
	fs=fopen("fflcvpcvq.txt", "a");
	ft=fopen("fflhisto.txt", "w");
	double *vetn;
	double *vetdet;
	double *vetdet1;
	vetn=(double*)calloc(6,sizeof(double));
	vetdet=(double*)calloc(6,sizeof(double));
	vetdet1=(double*)calloc(6,sizeof(double));
	
	double nMTFss=(bMTF/dMTF)*convfact;
	double nTFss=(tTF/dTF)*nMTFss;
	double nmuss=(tmu/dmu)*nTFss;
	double nMPss=tMP*nTFss/(dMP+a*tmu/dmu*kp/(km+a)*nTFss/convfact);
	double npss=(tp/dp)*nMPss;
	double ncss=(kp/(km+a))*(nMPss*nmuss)/convfact;
	double phi;
	
	start=time(NULL);
	
	phi=nMTFss/(double)convfact;
    phip=npss/(double)convfact;
    /*deterministic concentrations for the given set of parameters -- to be used only as check*/
	printf("phiMTF=%lf, nMTFss=%lf, nTFss=%lf, nmuss=%lf, nMPss=%lf, ncss=%lf, npss=%lf, phip=%lf\n", phi, nMTFss, nTFss, nmuss, nMPss, ncss, npss, phip);
	
    
	/*initial conditions*/
	vetn[0]=nMTFss;
	vetn[1]=nTFss;
	vetn[2]=nmuss;
	vetn[3]=nMPss;
	vetn[4]=ncss;
	vetn[5]=npss;
	vetdet[0]=vetn[0];
	vetdet[1]=vetn[1];
	vetdet[2]=vetn[2];
	vetdet[3]=vetn[3];
	vetdet[4]=vetn[4];
	vetdet[5]=vetn[5];
	vetdet1[0]=0.;
	vetdet1[1]=0.;
	vetdet1[2]=0.;
	vetdet1[3]=0.;
	vetdet1[4]=0.;
	vetdet1[5]=0.;

	
	fprintf(fp, "%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\n", t, vetn[0], vetn[1], vetn[2], vetn[3], vetn[4], vetn[5]);	
	
	while(t<pow(10., 7.)){
		t+=onestepgillespie(vetn);
		//printf("%.1lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\n", t, vetn[0], vetn[1], vetn[2], vetn[3], vetn[4], vetn[5]);
		sum=vetn[0]+vetn[1]+vetn[2]+vetn[3]+vetn[4]+vetn[5];
        
        if (sum>N){
            printf("Number of molecules too high,  N_TOT=%lf\n", sum);
            break;
        }
        
               
		tnew=(int)(t/step);
		
		//if(t>pow(10.,3.)) printf("%lf\n",tnew);
		
		if(tnew>=told+1){
		fprintf(fp, "%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\t%.3lf\n", t, vetn[0], vetn[1], vetn[2], vetn[3], vetn[4], vetn[5]);
		fprintf(ft, "%.3lf\n", vetn[5]);
		told=tnew;
		}
		
		if(t>pow(10.,3.)){
		tau++;
		sumsqp+=(vetn[5]*vetn[5]);
		sumsqTF+=(vetn[1]*vetn[1]);
		sump+=(vetn[5]);
		sumTF+=(vetn[1]);
        sumMTF+=(vetn[0]);
        summu+=(vetn[2]);
        sumMP+=(vetn[3]);
        sumc+=(vetn[4]);
        }
		i=i+1;
		
	}
	printf("Averaged for: %ld\n",tau);
	averp=sump/(double)tau;
	averTF=sumTF/(double)tau;
    averMTF=sumMTF/(double)tau;
    avermu=summu/(double)tau;
    averc=sumc/(double)tau;
    averMP=sumMP/(double)tau;
    varTF=(sumsqTF/(double)tau);
	varp=(sumsqp/(double)tau);
	xiTF=(varTF-pow(averTF,2.));
	xip=(varp-pow(averp,2.));
	
	fprintf(fs, "%lf       %lf    %lf     %lf    %lf        %lf          %.9lf     %lf\n", xip, xiTF, averp, averTF, varTF, varp, kp, convfact);
	printf("%lf\t%lf\t%lf\t%lf\t%lf\n", averMTF, avermu, averMP, averc,averp);
    
	end=time(NULL);
	
	elapsed_time=(int)(end-start);
	h=(int)elapsed_time/3600;
	min=(int)((elapsed_time-h*3600)/60);
	sec=(elapsed_time-h*3600-min*60);
	
	printf("\nElapsed time  %d h  %d min  %d sec \n \n",h, min, sec); 
	printf("\a");
	
	fclose(fp);
	fclose(fs);
	fclose(fp);
	free(vetn);
	free(vetdet);
	free(vetdet1);
	return(0);
}

/*function for the Gillespie's evolution*/
double onestepgillespie(double *vetn){
	double a0=0., a1=0., a2=0., a3=0., a4=0., a5=0., a6=0., a7=0., a8=0., a9=0., a10=0., a11=0., a12=0., a13=0., a14=0.;
	double p, tau, r1, r2;
	int check, N1=0;
	
	p=0.;
	
	
	/*calculus of reactions rate*/
	a1=bMTF*convfact;
	a2=dMTF*vetn[0];
	a3=tTF*vetn[0];
	a4=dTF*vetn[1];
	a5=tmu*vetn[1];
	a6=dmu*vetn[2];
	a7=tMP*vetn[1];
	a8=dMP*vetn[3];
	a9=kp*vetn[3]*vetn[2]/convfact;
	a10=km*vetn[4];
	a11=dc*vetn[4];
	a12=a*vetn[4];
	a13=dp*vetn[5];
	a14=tp*vetn[3];

	a0=(a1+a2+a3+a4+a5+a6+a7+a8+a9+a10+a11+a12+a13+a14);
	
	
	/*beginning of gillespie's step*/
	r1=drand48();
	r2=drand48();
	
	tau=(-1./a0)*log(r1);
	r2=a0*r2;
	
	check=0;
	
	
	
	while(check==0){
		
		
		p=a1;

		if (r2<p){
			vetn[0]=vetn[0]+1.;
			check=1;
			break;
		}
		
		p+=a2;
		
		if (r2<p && vetn[0]>=1.){
			vetn[0]=vetn[0]-1.;
			check=1;
			break;
		}
		
		p+=a3;
		
		if (r2<p){
			vetn[1]=vetn[1]+1.;
			check=1;
			break;
		}
		
		p+=a4;
		
		if (r2<p && vetn[1]>=1.){
			vetn[1]=vetn[1]-1.;
			check=1;
			break;
		}
		
		p+=a5;
		
		if (r2<p){
			vetn[2]=vetn[2]+1.;
			check=1;
			break;
		}
		
		p+=a6;
		
		if (r2<p && vetn[2]>=1.){
			vetn[2]=vetn[2]-1.;
			check=1;
			break;
		}
		
		p+=a7;
		
		if (r2<p){
			vetn[3]=vetn[3]+1.;
			check=1;
			break;
		}
		
		p+=a8;
		
		if (r2<p && vetn[3]>=1.){
			vetn[3]=vetn[3]-1.;
			check=1;
			break;
		}
		
		p+=a9;
		
		if (r2<p && vetn[3]>=1. && vetn[2]>=1.){
			vetn[4]=vetn[4]+1.;
			vetn[3]=vetn[3]-1.;
			vetn[2]=vetn[2]-1.;
			check=1;
			break;
		}
		
		p+=a10;
		
		if (r2<p && vetn[4]>=1.){
			vetn[2]=vetn[2]+1.;
			vetn[3]=vetn[3]+1.;
			vetn[4]=vetn[4]-1.;
			check=1;
			break;
		}
		
		p+=a11;
		
		if (r2<p && vetn[4]>=1.){
			vetn[4]=vetn[4]-1.;
			check=1;
			break;
		}
		
		p+=a12;
		
		if (r2<p && vetn[4]>=1.){
			vetn[4]=vetn[4]-1.;
			vetn[2]=vetn[2]+1.;
			check=1;
			break;
		}
		
		p+=a13;
		
		if (r2<p && vetn[5]>=1.){
			vetn[5]=vetn[5]-1.;
			check=1;
			break;
		}
		
		p+=a14;
		
		if (r2<p){
			vetn[5]=vetn[5]+1.;
			check=1;
			break;
		}
		
		
		
	}
	
	return(tau);
}










