// This program performs the stochastic simulation for the GFFL varying \theta_\mu and \theta_{mp}

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

/* Functions */
double gillespie_step(double *RATES, int *STNEW, int *STOLD);
void inizialize_state_vectors(int TF0, int TFA0, int MU0, int MP0, int C0, int P0, int PA0, int *STOLD);
void fill_in_with_zeros_double(int ARRAY_DIM, double *ARRAY_TO_FILL_IN);


int main()
{
  double thetaTF, deltaTF;                                                                        /* Rates of the chemical equations for TF                                 */
  double thetaMU, deltaMU;                                                                        /* Rates of the chemical equations for miRNA                              */
  double thetaMP, deltaMP;                                                                        /* Rates of the chemical equations for mRNAP                              */
  double kCp, kCm;                                                                                /* Binding and unbinding rates of microRNA with mRNA_P                    */
  double thetaP, deltaP;                                                                          /* Rates of the chemical equations for P                                  */
  double kRp, kRm;                                                                                /* Binding and unbinding rates between P and DNA free                     */
  double kTFp, kTFm;                                                                              /* Binding and unbinding rates between DNA and the enzyme for mRNA_TF     */
  double alpha;
  double st_t_old, st_t_new;                                                                      /* Stochastic time vector                                                 */
  double *rates;                                                                                  /* Vector containing all the rates                                        */                            
  int nTF0, nTFa0, nMU0, nMP0, nC0, nP0, nPa0;                                                    /* Initial number of molecules                                            */
  int *st_n_new, *st_n_old;                                                                       /* Stochastic state vectors                                               */
  int steps;                                                                                      /* Number of Gillespie steps                                              */
                                                                                                  /* Lattice dimension                                                      */
  double N;
  double fattConv;
  int a, averages;
  int skip, skip_termalization, s;
  double meanTFsim, meanTFasim, meanMUsim, meanMPsim, meanCsim, meanPsim, meanPasim;
    
  int bialek, selfRepression, miRNA;
  double sigmaPnold, sigmaPn;
  double st_t_start;

  double dim;

  long double sigmaNew, sigmaNewOld, meanNew;
  double indexKTF;
  
  int i,j, indexMethod;
  double  indexMU;
  int noise;
 
  time_t start, end;
  int elapsed_time, h, min, sec; 
  FILE *file;
  char file_name[150];

  

  averages = 1;

    steps = 100000000;

    skip_termalization = 3000;
  
  skip = 1000;
  N=1e6;
  
  fattConv=2.52e3;
  
  start = time(NULL);

  for(indexMethod=0; indexMethod<3; indexMethod++){
    if(indexMethod==0){
      bialek=1;
      selfRepression=0;
      miRNA=0;
    }
    if(indexMethod==1){
      bialek=0;
      selfRepression=1;
      miRNA=0;
    }
    if(indexMethod==2){
      bialek=0;
      selfRepression=0;
      miRNA=1;
    }
    
    /* Rates */
    thetaTF = 0.01*fattConv;  
    //thetaTF = 0.4;  
    thetaMP = 0.0798071*fattConv;
    deltaMP = 0.0243291;      
    thetaP =  0.0271314;
    deltaP =  0.01;
    deltaTF = deltaP; 
    kTFp = 0.8;
    kTFm = 200.0;
    
    if(bialek==1){
      kCp = 0.0;
      kCm = 0.0;
      thetaMU = 0.0;
      deltaMU = 0.0;
      kRp = 0.0;
      kRm = 0.0;
      alpha=0.0;
    }
    if(miRNA==1){
      thetaMU=0.0798071;
      deltaMU=0.0154351;
      kCp=0.01;
      kCm = 0.0001;
      kRp = 0.0;
      kRm = 0.0;
      alpha=0.001;
    }
    if(selfRepression==1){
      kCp = 0.0;
      kCm = 0.0;
      thetaMU = 0.0;
      deltaMU = 0.0;
      
      kRp = 100;
      kRm = 0.01;
      alpha=0;
    }
    
    /* Initial values */
    nTF0   = 1000;
    nTFa0  = 1;
    nMU0   = 1000;
    nMP0   = 300;
    nP0    = 500;
    nC0    = 2000;
    nPa0   = 1;
    
    meanTFsim  = 0.0;
    meanTFasim = 0.0;
    meanMUsim  = 0.0;
    meanMPsim  = 0.0;
    meanCsim   = 0.0;
    meanPsim   = 0.0;
    meanPasim  = 0.0;
    
    sigmaPn = 0.0;
    
    /* Initialization of arrays */
    st_n_new   = (int *) calloc(7,sizeof(int));
    st_n_old   = (int *) calloc(7,sizeof(int));
    rates    = (double *) calloc(16,sizeof(double));
    
    /* Save rates in the array */
    rates[0]  = thetaTF;
    rates[1]  = kTFp;
    rates[2]  = kTFm;
    rates[3]  = deltaTF;
    rates[4]  = thetaMU;
    rates[5]  = deltaMU;
    rates[6]  = thetaMP;
    rates[7]  = kRp;
    rates[8]  = kRm;
    rates[9]  = deltaMP; 
    rates[10] = kCp;
    rates[11] = kCm;
    rates[12] = thetaP;
    rates[13] = deltaP;
    rates[14] = N;
    rates[15] = alpha;
    
    
    
    sprintf(file_name,"sim_steps_%2.1e_A_%3.2e_tMU_%3.2e_kCp_%3.2e_kCm_%3.2e_kRp_%3.2e_kRm_%3.2e_alpha_%3.2e.dat", (double)steps, (double)averages, thetaMU, kCp, kCm, kRp, kRm, alpha);
    file=fopen(file_name,"w");
    
    for(thetaTF=1000.01; thetaTF< 1100.1; thetaTF+= 10000.1){
      
      rates[0] = thetaTF;
      
	for(indexMU=-5.0; indexMU< -2.0; indexMU+= 0.1){
	 thetaMU = pow(10.0, indexMU);

	rates[4]  = thetaMU;
	
	if(bialek==1){
	  rates[4] = 0.0;
	}
	
	if(selfRepression==1){
	  rates[4] = 0.0;  
	}
	
	for(thetaMP=0.02; thetaMP< 1.0; thetaMP+= 0.02){
	  rates[6]  = thetaMP*fattConv;
	  
	  sigmaPn = 0.0;
	  sigmaNew=0.0;
	  
	  
	  for(a=0; a<averages; a++){
	    /* Initialization  */
	    inizialize_state_vectors(nTF0, nTFa0, nMU0, nMP0, nC0, nP0, nPa0, st_n_old);
	    st_t_old = 0.0;
	    sigmaPnold = 0.0;
	    meanPsim   = 0.0;
	    
	    sigmaNewOld=0.0;
	    meanNew=0.0;
	    
	    /* FFL */
	    s=1;
	    dim=0.0;
	    while(s<steps){  
	      st_t_new = st_t_old + gillespie_step(rates, st_n_new, st_n_old);
   
	      if(st_n_new[0] +st_n_new[1] +st_n_new[2] +st_n_new[3] +st_n_new[4] +st_n_new[5] +st_n_new[6] >(int) N){
	      
		printf("Too many molecules for thetaTF %6.4e and ktfp %6.4e\n",thetaTF, kTFp);
		s=steps;
		break;
	      }
	      if(s==skip_termalization){
		st_t_start=st_t_new;
	      }  
	      if(s>skip_termalization){
		if( (double)st_n_old[5] >0.0){
		  meanPsim   +=( (double)st_n_old[5]*(st_t_new-st_t_old)); 
		  sigmaPnold += ((double)st_n_old[5]*(double)st_n_old[5]  *(st_t_new-st_t_old));
		  meanNew   +=(long double)st_n_old[5]; 
		  sigmaNewOld += ((long double)st_n_old[5]*(long double)st_n_old[5]);
		  dim+=1.0;
		  
		}
		
	    }
	      s++;
	      st_t_old = st_t_new; 
	      st_n_old[0] = st_n_new[0];
	      st_n_old[1] = st_n_new[1];
	      st_n_old[2] = st_n_new[2];
	      st_n_old[3] = st_n_new[3];
	      st_n_old[4] = st_n_new[4];
	      st_n_old[5] = st_n_new[5];
	      st_n_old[6] = st_n_new[6];
	    }
	    meanPsim   /= (st_t_new-st_t_start);
	    sigmaPnold /= (st_t_new-st_t_start);
	    
	    sigmaPnold = sqrt(sigmaPnold-meanPsim*meanPsim)/meanPsim;
	    sigmaPn += sigmaPnold;
	    
	    
	    sigmaNewOld=sqrt(dim*sigmaNewOld-meanNew*meanNew)/meanNew;
	    sigmaNew+=sigmaNewOld;
	  }
	  
	  sigmaPn /=(double)averages;
	  sigmaNew /=(long double)averages;
	  
	  fprintf(file,"%6.4e %6.4e %6.4e  %6.4e %6.4Le \n", rates[0], rates[4],  thetaMP, sigmaPn, sigmaNew);	
	 
	} 
      }
    }
    
    fclose(file);
    
    /* Free arrays */
    free(st_n_old);
    free(st_n_new);
    free(rates);
  }
  
  /* Elapsed time */
  end = time(NULL);
  elapsed_time = (int)(end-start);
  h=(int)(elapsed_time/3600);
  min=(int)((elapsed_time-h*3600)/60);
  sec=elapsed_time-h*3600-min*60;
  printf("\nElapsed time  %d h  %d min  %d sec \n \n",h, min, sec); 
  
}



/*******************************************************************************/
/*                                  Functions                                  */
/*******************************************************************************/

void inizialize_state_vectors(int TF0, int TFA0, int MU0, int MP0, int C0, int P0,int PA0, int *STOLD){
  STOLD[0] = TF0;
  STOLD[1] = TFA0;
  STOLD[2] = MU0;
  STOLD[3] = MP0;
  STOLD[4] = C0;
  STOLD[5] = P0;
  STOLD[6] = PA0; 
}


void fill_in_with_zeros_double(int ARRAY_DIM, double *ARRAY_TO_FILL_IN){
  int i;
  for(i=0; i<ARRAY_DIM; i++){
    ARRAY_TO_FILL_IN[i]=0.0;
  }
}


double gillespie_step(double *RATES, int *STNEW, int *STOLD){
  double tau, prob;
  double r1, r2;
  double a0, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15;
  double THETATF, KTFP, KTFM, DELTATF, THETAMU, DELTAMU, THETAMP, KRP, KRM, DELTAMP, KCP, KCM, THETAP, DELTAP, NN, ALPHA;
  
  int index;
  int ind;
  

  THETATF = RATES[0];
  KTFP    = RATES[1];
  KTFM    = RATES[2];
  DELTATF = RATES[3];
  THETAMU = RATES[4];
  DELTAMU = RATES[5];
  THETAMP = RATES[6];
  KRP     = RATES[7];
  KRM     = RATES[8];
  DELTAMP = RATES[9]; 
  KCP     = RATES[10];
  KCM     = RATES[11];
  THETAP  = RATES[12];
  DELTAP  = RATES[13];

  NN      = RATES[14];  

  ALPHA   = RATES[15];  

  
  tau=0.0;
  prob=0.0;
  ind=0;
  r1=0.0;
  r2=0.0;
  a0=0.0;

  
  /* Transition rates */
  if(STOLD[1]==1){
    a1  = THETATF/NN;
    a2  = 0.0;
    if(KTFM<1e-8){
      a3=0.0;
    }
    else{
      a3  = KTFM/NN;  
    }
  }
  else{
    a1  = 0.0;
    if(KTFP<1e-8){
      a2=0.0;
    }
    else{
      a2  = KTFP/NN;
    }
    a3  = 0.0;
  }
  
  a4  = DELTATF * (double)STOLD[0]/NN;
  a5  = THETAMU * (double)STOLD[0]/NN;
  a6  = DELTAMU * (double)STOLD[2]/NN;
  a7  = THETAMP * (double)STOLD[0]/NN * (double)STOLD[6]/NN;
  a8  = KRP * (double)STOLD[6]/NN * (double)STOLD[5]/NN;
  a9  = KRM * (1.0 - (double)STOLD[6])/NN;
  a10 = DELTAMP * (double)STOLD[3]/NN;
  a11 = KCP * (double)STOLD[3]/NN * (double)STOLD[2]/NN;
  a12 = KCM * (double)STOLD[4]/NN;
  a13 = THETAP * (double)STOLD[3]/NN;
  a14 = DELTAP * (double)STOLD[5]/NN;
  a15 = ALPHA* (double)STOLD[4]/NN;
  
  a0 = a1+ a2+ a3+ a4+ a5+ a6+ a7+ a8+ a9+ a10+ a11+ a12+ a13+ a14+ a15;

  r1 = drand48();
  r2 = drand48();
  tau=-(1.0/a0)*(log(r1));
  r2=a0*r2;
  
  STNEW[0] = STOLD[0];
  STNEW[1] = STOLD[1];
  STNEW[2] = STOLD[2];
  STNEW[3] = STOLD[3];
  STNEW[4] = STOLD[4];
  STNEW[5] = STOLD[5];
  STNEW[6] = STOLD[6];

  while(ind==0){
    prob=prob+a1;
    if(r2<prob){
      STNEW[0]+=1;
      ind=1;
      break;
    }
    prob=prob+a2;
    if(r2<prob){
      STNEW[1]+=1;
      ind=1;
      break;  
    }
    prob=prob+a3;
    if(r2<prob){
      STNEW[1]-=1;
      ind=1;
      break;
    }
    prob=prob+a4;
    if(r2<prob){
      STNEW[0]-=1;
      ind=1;
      break;
    }
    prob=prob+a5;
    if(r2<prob){
      STNEW[2]+=1;
      ind=1;
      break;
    }
    prob=prob+a6;
    if(r2<prob){
      STNEW[2]-=1;
      ind=1;
      break;
    }
    prob=prob+a7;
    if(r2<prob){
      STNEW[3]+=1;
      ind=1;
      break;
    }
    prob=prob+a8;
    if(r2<prob){
      STNEW[6]-=1;
      STNEW[5]-=1;
      ind=1;
      break;
    }
    prob=prob+a9;
    if(r2<prob){
      STNEW[6]+=1;
      STNEW[5]+=1;
      ind=1;
      break;
    }
    prob=prob+a10;
    if(r2<prob){
      STNEW[3]-=1;
      ind=1;
      break;
    }
    prob=prob+a11;
    if(r2<prob){
      STNEW[3]-=1;
      STNEW[2]-=1;
      STNEW[4]+=1;
      ind=1;
      break;
    }
    prob=prob+a12;
    if(r2<prob){
      STNEW[3]+=1;
      STNEW[2]+=1;
      STNEW[4]-=1;
      ind=1;
      break;
    }
    prob=prob+a13;
    if(r2<prob){
      STNEW[5]+=1;
      ind=1;
      break;
    }
    prob=prob+a14;
    if(r2<prob){
      STNEW[5]-=1;
      ind=1;
      break;
    }
    prob=prob+a15;
    if(r2<prob){
      STNEW[4]-=1;
      STNEW[2]+=1;
      ind=1;
      break;
    }
    
    if(ind==0){
      ind=1;
      printf("Error in Gillespie !!!!! \n");
    }   
  }

  return(tau);
  
} 







